# -*- coding: utf-8 -*-
"""
Created on Sun Sep  5 21:10:20 2021

@author: a
"""
import torch
import torch.nn as nn
import random

#device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class Encoder(nn.Module):
    def __init__(self, inp_dim, hid_dim, bsz=64, n_layers=1):
        super(Encoder, self).__init__()
        h0 = torch.zeros(n_layers, bsz, hid_dim)
        c0 = torch.zeros(n_layers, bsz, hid_dim)
        nn.init.xavier_normal_(h0, gain=nn.init.calculate_gain('relu'))
        nn.init.xavier_normal_(c0, gain=nn.init.calculate_gain('relu'))
        self.h0 = h0
        self.c0 = c0
        self.LSTM = nn.LSTM(input_size=inp_dim,
                            hidden_size=hid_dim,
                            num_layers=n_layers)
    def forward(self, x):
        outputs, (hidden, cell) = self.LSTM(x, (self.h0, self.c0))
        return hidden, cell

class Decoder(nn.Module):
    def __init__(self, out_dim, hid_dim, bsz=64, n_layers=1):
        super(Decoder, self).__init__()
        h0 = torch.zeros(n_layers, bsz, hid_dim)
        c0 = torch.zeros(n_layers, bsz, hid_dim)
        nn.init.xavier_normal_(h0, gain=nn.init.calculate_gain('relu'))
        nn.init.xavier_normal_(c0, gain=nn.init.calculate_gain('relu'))
        self.h0 = nn.Parameter(h0, requires_grad=True)
        self.c0 = nn.Parameter(c0, requires_grad=True)
        self.LSTM = nn.LSTM(input_size=out_dim,
                            hidden_size=hid_dim,
                            num_layers=n_layers)
    def forward(self, x, hidden, cell):
        output, (hidden, cell) = self.LSTM(x, (hidden, cell))
        return output, (hidden, cell)

class LSTM_AE(nn.Module):
    def __init__(self,
                 inp_dim,
                 out_dim,
                 hid_dim,
                 code_size,
                 bsz=64,
                 tf_p = 0.5,
                 n_layers=1):
        super(LSTM_AE, self).__init__()
        self.bsz = bsz
        self.input_dim = inp_dim
        self.output_dim = out_dim
        self.hidden_dim = hid_dim
        self.code_size = code_size
        self.tf_p = 0.9    # teacher forcing 
        
        self.encoder = Encoder(hid_dim, hid_dim, bsz, n_layers)
        self.decoder = Decoder(hid_dim, hid_dim, bsz, n_layers)
        
        self.fc_encoder1 = nn.Linear(self.input_dim, self.hidden_dim)
        self.fc_decoder1 = nn.Linear(self.hidden_dim, self.output_dim)
        
        self.fc_encoder2 = nn.Linear(self.hidden_dim, self.code_size)
        self.fc_decoder2 = nn.Linear(self.code_size, self.hidden_dim)
        
        self.fc_sigmoid = nn.Sigmoid()
        self.out_sigmoid = nn.Sigmoid()

    def forward(self, src):
        src = src.contiguous().view(len(src), self.input_dim, -1)    # shape: (batch_size, inp_dim, seq_len)
        src = src.transpose(1, 2)    # transpose dim0 and dim1  shape: (batch_size, seq_len, inp_dim)        
        x = self.fc_sigmoid(self.fc_encoder1(src))    # shape: (batch_size, seq_len, code_dim)        
        x = x.transpose(0, 1)    # transpose dim0 and dim1  shape: (seq_len, batch_size, inp_dim)
        hidden, cell = self.encoder(x)    # hidden, cell: (layer_nums, batch_size, code_dim)
          
        hidden = self.fc_sigmoid(self.fc_encoder2(hidden[0]))    # shape: (batch_size, code_dim)
        encoded_ = hidden.clone().detach()
        
        hidden = self.fc_sigmoid(self.fc_decoder2(hidden))    #  shape: (batch_size, output_dim)
        
        hidden = hidden.unsqueeze(0)    # shape: (layer_nums, batch_size, output_dim)
        seq_len = x.shape[0] # seq_len
        outputs = torch.zeros(seq_len, self.bsz, self.hidden_dim)
        # start first input
        input_ = x[0:1,:]     # input_: (1, batch_size, code_dim)
        for t in range(1, seq_len): # 1 to seq_len-1, except first input
            output, (hidden, cell) = self.decoder(input_, hidden, cell)
            outputs[t] = output[0]
            
            if random.random() < self.tf_p:     # teacher forcing mechanism
                input_ = x[t:t+1, :]
            else:
                input_ = output[0:1, :]
                
        outputs = torch.transpose(outputs, 0, 1)    # transpose dim0 and dim1    shape: (batch_size, seq_len, code_size)
        outputs = self.out_sigmoid(self.fc_decoder1(outputs))    #  shape: (batch_size, seq_len, output_dim)
        outputs = torch.transpose(outputs, 1, 2)    # transpose dim0 and dim1    shape: (batch_size, out_dim, seq_len)
        outputs = outputs.contiguous() .view(len(outputs), 1, -1)  # transpose dim0 and dim1    shape: (batch_size, out_dim, seq_len)]
        return encoded_, outputs