# -*- coding: utf-8 -*-
"""
Created on Sun Sep  5 16:29:39 2021

@author: a
"""

import torch
import torch.nn as nn
import torch.utils.data as Data
import numpy as np
from model1 import LSTM_AE
#from model2 import LSTMAE
from sklearn import cluster, metrics
import math
import matplotlib.pyplot as plt

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class phone_data(Data.Dataset):
    def __init__(self,path, normalization = 'Two'):
        self.normalization = normalization
        self.phone = self.load(path)
    
    def __getitem__(self, index):
        return self.phone[index]
    
    def load(self,path):
        train_data01 = []
        train_data02 = []
        with open(path,'r+') as f:
            for line in f:
                pData = line.strip().split(",")
                pData = np.reshape(pData, (1, 168)).astype(float)
                p_max = np.max(pData)
                train_data01.append(pData/(p_max+10e-7))
                train_data02.append(pData)
        train_data01 = np.array(train_data01)
        train_data02 = np.array(train_data02)
        zmax_train02 = train_data02.max(axis = 0)
        train_data02 = train_data02/(zmax_train02+10e-7)
        train_data = np.concatenate((train_data01, train_data02), axis = -1)
        return train_data
    
    def __len__(self):
        return len(self.phone)

if __name__ == '__main__':
    # Hyper Parameters
    EPOCH = 100
    BATCH_SIZE = 16
    LR = 0.005         # learning rate
    LATENT_CODE_NUM = 7
    k = 10    # the expected speed of convergence
    
    data = phone_data(r'./data/su_jz for cluster2.csv')
    train_loader = Data.DataLoader(dataset    = data,
                                   batch_size = BATCH_SIZE,
                                   shuffle    = True,
                                   drop_last  = True)
    train_loader1 = Data.DataLoader(dataset    = data,
                                   batch_size = 1,
                                   shuffle    = False)
    
#    model = LstmFcAutoEncoder(input_layer  = 336,
#                              hidden_layer = 100,
#                              code_size    = LATENT_CODE_NUM,
#                              batch_size   = BATCH_SIZE)  # lstm+fc1       
    model = LSTM_AE(inp_dim   = 2,
                    out_dim   = 2, 
                    hid_dim   = 48,
                    code_size = LATENT_CODE_NUM,
                    bsz       = BATCH_SIZE,
                    tf_p      = 1,    # Teacher forcing with Scheduled Sampling
                    n_layers  = 1)
#    model = LSTMAE(input_size     = 1,
#                    hidden_size   = LATENT_CODE_NUM,
#                    dropout_ratio = 0.5,
#                    seq_len       = 336)
    
    loss_function = nn.BCELoss(reduction = 'sum')  # loss
    optimizer = torch.optim.Adam(model.parameters(), lr=LR)  # 优化器
    
    # 开始训练
    model.train()
    loss_hist = []
    for i in range(EPOCH):
        total_loss = 0
        for p_seq in train_loader:
            seq = p_seq.float()
            _, y_pred = model(seq)
            single_loss = loss_function(y_pred[:,:,1:], seq[:,:,1:])
            
            optimizer.zero_grad()
            single_loss.backward()
            optimizer.step()
            
            total_loss += single_loss.item()
        print('====> Epoch: {} Average loss: {:.6f}'.format(
            i, total_loss / len(train_loader.dataset)))
        loss_hist.append(total_loss / len(train_loader.dataset))
        if i % 20 == 0:
            test_data = torch.tensor(data[:BATCH_SIZE]).float()    # top-32 used for testing
            _, y_pred = model(test_data)
            print("       Test LOSS: ", loss_function(y_pred, test_data).item()/BATCH_SIZE)
        
        model.tf_p = max(k / (k + math.exp((i - 10) / k)), 0.4)    # Scheduled sampling : Inverse sigmoid decay
    
    torch.save(model.state_dict(), './trained model/model_ep100_hidden7.pth')
    
    #  encode
    model.eval()
    encoded_data = []
    decoded_data = []
    for seq in train_loader1:
        seq = seq.float()
        seq = seq.repeat(BATCH_SIZE, 1, 1)
        encode, decode = model(seq)
        encoded_data.append(encode[0,:].contiguous().view((-1)).detach().numpy())
        decoded_data.append(decode[0,:].contiguous().view((-1)).detach().numpy())
    encoded_data = np.array(encoded_data)
    decoded_data = np.array(decoded_data)
    print(encoded_data.shape)
    
    # kmeans
    nclusters = 5
    kmeans = cluster.KMeans(n_clusters = nclusters)
    kmeans.fit(encoded_data)
    y_cls = kmeans.labels_
    y_centers = kmeans.cluster_centers_
    np.savetxt('1_cls.csv', y_cls,  fmt="%s", delimiter = ',')
    np.savetxt('1_center.csv', y_centers,  fmt="%s", delimiter = ',')
    
    score = metrics.calinski_harabaz_score(encoded_data, y_cls)
    score1 = metrics.silhouette_score(encoded_data, y_cls)
    print("method1 clusters value is %f\t%f\n" % (score, score1))
    
    # kmeans with diffence n_clusters
    for n in range(2,20):
        kmeans = cluster.KMeans(n_clusters = n)
        kmeans.fit(encoded_data)
        score = metrics.calinski_harabaz_score(encoded_data,kmeans.labels_)
        score1 = metrics.silhouette_score(encoded_data,kmeans.labels_)
        print("%d clusters kmeans value is %f\t%f" % (n, score, score1))